/*
 * pqboost.c -- An add-on to boost the PHOST build queue (PDK version of QBOOST)
 *
 * Copyright (C) 1998-2002 Stefan Glasauer
 *
 * Author: Stefan Glasauer <S.Glasauer@lrz.uni-muenchen.de>
 * PDK version by Heiko Schlenker <hschlen@gmx.de>
 * Version: $Revision: 1.4 $
 * Status: quick & dirty
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Id: pqboost.c,v 1.4 2002/05/29 17:34:56 hschlen Exp $
 */

/*
 * $Log: pqboost.c,v $
 * Revision 1.4  2002/05/29 17:34:56  hschlen
 * New option EnableESBandFCB, support for non-standard race assignments
 *
 * Revision 1.3  2001/03/06 21:29:23  hschlen
 * *** empty log message ***
 *
 * Revision 1.2  2001/03/06 20:40:56  hschlen
 * added replacement snprintf() function
 *
 * Revision 1.1.1.1  2001/02/01 20:39:39  hschlen
 * initial import into CVS
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <regex.h>
#include <time.h>
#include <phostpdk.h>

#ifdef __sun__
extern char *sys_errlist[];
#endif

#define OLD_RACENR 11

static const char rcsid[] = "$Id: pqboost.c,v 1.4 2002/05/29 17:34:56 hschlen Exp $";
const char PQBOOST_VERSION[] = "2.0.4";
const char ReleaseDate[] = "$Date: 2002/05/29 17:34:56 $";
const char Notice[] = "Copyright (C) 1998-2002 Stefan Glasauer.\n\
PDK version by Heiko Schlenker.\n\
PQBOOST comes with ABSOLUTELY NO WARRANTY.\n\
This is free software. See COPYING for details.";

#define ERRLEN  256     /* max lemgth of error message */
#define SUBEXPR 4       /* number of sub expressions */

#define MAXLINE 1000    /* maximum characters per line read by fgets() */
#define MAXMSGLINE 42   /* A VGAP message is assumed to be a plain-text
                           null-terminated string not exceeding 840 characters
                           (20 lines by 42 characters, with a 0Dh at the end of
                           each line). */

typedef enum {
    noranking=0,
    individual=1,
    public=2,
    verbosepublic=3
} RankingList_Def;


/* PQBOOST configuration file structure */

struct pqboostconfig_struct {
    Uns16 beamsrequired;
    Uns16 baysrequired;
    double relativepower;
    Boolean enablesecurity;
    Boolean enableesbandfcb;
    Boolean alliancegamemode;
    double shipfactor;
    double planetfactor;
    double minefieldfactor;
    double combatfactor;
    double visionfactor;
    RankingList_Def showranking;
    Boolean showtotalbattlemass;
    Uns16 rankinglistinterval;
};

struct pqboostconfig_struct pqboostconfig = {
    5, 3, 2.00, True, False, False, 0.4, 0.2, 0.3, 0.4, 0.2, noranking, False, 1
};


/* Define an array of players. */

struct player {
    Uns16 id;
    Int32 battlemass;
    Int32 pal;
    double alliespalportion[RACE_NR + 1];  /* race 0 is nothing */
    Uns16 ranking;
};

struct player playerslist[] = {
    {0, 0, 0, {0.0}, 0},
    {1, 0, 0, {0.0}, 0},
    {2, 0, 0, {0.0}, 0},
    {3, 0, 0, {0.0}, 0},
    {4, 0, 0, {0.0}, 0},
    {5, 0, 0, {0.0}, 0},
    {6, 0, 0, {0.0}, 0},
    {7, 0, 0, {0.0}, 0},
    {8, 0, 0, {0.0}, 0},
    {9, 0, 0, {0.0}, 0},
    {10, 0, 0, {0.0}, 0},
    {11, 0, 0, {0.0}, 0}
};

int count = sizeof(playerslist) / sizeof(struct player);


/* These are the comparison functions used for sorting and searching. */

int playerfind_cmp(const void *p1, const void *p2)
{
    struct player *sp1 = (struct player *) p1;
    struct player *sp2 = (struct player *) p2;

    return (int) ((sp1->id) - (sp2->id));
}

int player_cmp(const void *p1, const void *p2)
{
    struct player *sp1 = (struct player *) p1;
    struct player *sp2 = (struct player *) p2;

    if ((sp2->battlemass) == (sp1->battlemass)) {
	return (int) ((sp2->id) - (sp1->id));
    } else {
	return (int) ((sp2->battlemass) - (sp1->battlemass));
    }
}


/* Do the lookup into the sorted array. */

struct player *find_player(const Int32 what, const Boolean useid)
{
    struct player target, *result;

    result = NULL;
    if (True == useid) {
	target.id = (Uns16) what;
	result = bsearch(&target, playerslist, count, sizeof(struct player), playerfind_cmp);
    } else {
	target.battlemass = what;
	result = bsearch(&target, playerslist, count, sizeof(struct player), player_cmp);
    }

    return (result ? result : NULL);
}


/* For non-standard race assignments. */
Uns16 EffRaceNR(Uns16 pPlayer)
{
    return (pPlayer >= 1 && pPlayer <= OLD_RACENR)
	? gPconfigInfo->PlayerRace[pPlayer]
	: 0;
}


/* Assignment function for the PDK function ConfigFileReader("pqboost.cfg"). */

Boolean qboost_assign(const char *lhs, char *rhs, const char *lInputLine)
{
    Boolean returncode = True;


    if ((lhs != NULL) && (rhs != NULL)) {
	if (strcmp(lhs, "BeamsRequired") == 0) {
	    if (atoi(rhs) >= 0) {
		pqboostconfig.beamsrequired = (Uns16) atoi(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "BaysRequired") == 0) {
	    if (atoi(rhs) >= 0) {
		pqboostconfig.baysrequired = (Uns16) atoi(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "RelativePower") == 0) {
	    if (atof(rhs) >= 0.0) {
		pqboostconfig.relativepower = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "EnableESBandFCB") == 0) {
	    if (atoi(rhs) == 0) {
		pqboostconfig.enableesbandfcb = False;
	    } else if (atoi(rhs) == 1) {
		pqboostconfig.enableesbandfcb = True;
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "EnableSecurity") == 0) {
	    if (atoi(rhs) == 0) {
		pqboostconfig.enablesecurity = False;
	    } else if (atoi(rhs) == 1) {
		pqboostconfig.enablesecurity = True;
	    } else {
		returncode = False;
	    }
	}
	if ((strcmp(lhs, "AllianceGameMode") == 0) || (strcmp(lhs, "TeamGameMode") == 0)) {
	    if (atoi(rhs) == 0) {
		pqboostconfig.alliancegamemode = False;
	    } else if (atoi(rhs) == 1) {
		pqboostconfig.alliancegamemode = True;
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "ShipFactor") == 0) {
	    if ((atof(rhs) >= 0.0) && (atof(rhs) <= 1.0)) {
		pqboostconfig.shipfactor = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "PlanetFactor") == 0) {
	    if ((atof(rhs) >= 0.0) && (atof(rhs) <= 1.0)) {
		pqboostconfig.planetfactor = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "MinefieldFactor") == 0) {
	    if ((atof(rhs) >= 0.0) && (atof(rhs) <= 1.0)) {
		pqboostconfig.minefieldfactor = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "CombatFactor") == 0) {
	    if ((atof(rhs) >= 0.0) && (atof(rhs) <= 1.0)) {
		pqboostconfig.combatfactor = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "VisionFactor") == 0) {
	    if ((atof(rhs) >= 0.0) && (atof(rhs) <= 1.0)) {
		pqboostconfig.visionfactor = atof(rhs);
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "ShowRanking") == 0) {
	    if (atoi(rhs) == 0) {
		pqboostconfig.showranking = noranking;
	    } else if (atoi(rhs) == 1) {
		pqboostconfig.showranking = individual;
	    } else if (atoi(rhs) == 2) {
		pqboostconfig.showranking = public;
	    }
	     else if (atoi(rhs) == 3) {
	     pqboostconfig.showranking = verbosepublic;
	     }
   	    else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "ShowTotalBattlemass") == 0) {
	    if (atoi(rhs) == 0) {
		pqboostconfig.showtotalbattlemass = False;
	    } else if (atoi(rhs) == 1) {
		pqboostconfig.showtotalbattlemass = True;
	    } else {
		returncode = False;
	    }
	}
	if (strcmp(lhs, "RankinglistInterval") == 0) {
	    if ((atoi(rhs) > 0) && (atoi(rhs) <= 255)) {
		pqboostconfig.rankinglistinterval = (Uns16) atoi(rhs);
	    } else {
		returncode = False;
	    }
	}
    }

    return returncode;
}


/* Modify PALPlayerRate entry in pconfig.src */

Boolean modify_palplayerrate(const Int16 phase, FILE *errorlogfile, const Boolean enablesecurity)
{
    Boolean returncode = True;
    FILE *inputfile = NULL;
    FILE *outputfile = NULL;
    FILE *savefile = NULL;
    FILE *retrievefile = NULL;
    Uns16 race = 0;
    char line[MAXLINE];
    char line2[MAXLINE];
    Boolean useid = True;
    Boolean reg_t_freed = False;
    Boolean regex_found = False;
    const char *rex = "^[^#]*PALPlayerRate.*=.*[0-9]*";  /* search pattern */
    regex_t reg_t;                                       /* compiled regexp */
    regmatch_t rm_t[SUBEXPR];                            /* matches of subexpr. */
    int ret = 0;                                         /* return value */


    inputfile = OpenInputFile("pconfig.qbo", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    outputfile = OpenOutputFile("pconfig.src", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (1 == phase) {
	savefile = OpenOutputFile("pqboost.tmp", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    }
    if (3 == phase) {
	retrievefile = OpenInputFile("pqboost.tmp", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    }

    if (inputfile && outputfile) {
	while (fgets(line, MAXLINE, inputfile)) {
	    /* first compile regular expression for further use
	     * (see *POSIX Regexp Compilation*) with an useful flag
	     * (see *Flags for POSIX Regexps*)
	     */
	    ret = regcomp(&reg_t, rex, REG_EXTENDED);
	    if (ret != 0) {
		char errbuf[ERRLEN];
		regerror(ret, &reg_t, errbuf, ERRLEN);
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error compiling regular expression: %s\n",
			errbuf);
		returncode = False;
		goto quit;
	    }

	    /* now search if argv[1] matches the regexp and store the first
	     * SUBEXPR (#defined as 4) matching subexpressions in <rm_t>
	     * (actually, there is only 1 ;-) (see *Regexp Subexpressions*)
	     */
	    ret = regexec(&reg_t, line, SUBEXPR, rm_t, 0);
	    if (0 == ret) {
		/* if there was a match, write the new PALPlayerRate entry */
		regex_found = True;
		if (1 == phase) {
		    if (savefile) {
			/* save old PALPlayerRate */
			fprintf(savefile, "%s", line);
		    } else {
			fprintf(errorlogfile ? errorlogfile : stderr,
				"Error: Cannot create pqboost.tmp!\n");
			returncode = False;
			goto quit;
		    }
		}
		if (3 == phase) {
		    if (retrievefile) {
			if (True == enablesecurity) {
			    /* restore old PALPlayerRate */
			    if (fgets(line2, MAXLINE, retrievefile)) {
				fprintf(outputfile, "%s", line2);
				Info("Security: Old PALPlayerRate restored.");
			    } else {
				fprintf(errorlogfile ? errorlogfile :
					stderr,
					"Error: Cannot read contents of pqboost.tmp!\n");
				returncode = False;
				goto quit;
			    }
			} else {
			    Info("Security: Old PALPlayerRate NOT restored.");
			    fprintf(outputfile, "%s", line);
			}
		    } else {
			fprintf(errorlogfile ? errorlogfile : stderr,
				"Error: Could not open pqboost.tmp. Possible error: PQBOOST phase 1 not executed\n");
			returncode = False;
			goto quit;
		    }
		} else {
		    fprintf(outputfile, "PALPlayerRate = ");
		    for (race = 0; race <= RACE_NR; race++) {
			if (playerslist[race].id > 0) {
			    fprintf(outputfile, "%ld", find_player(playerslist[race].id, useid)->pal);
			    if (race < RACE_NR) {
				fprintf(outputfile, ",");
			    }
			}
		    }
		    fprintf(outputfile, "\n");
		}
	    } else if (REG_ESPACE == ret) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error: regexp ran out of memory\n");
	    } else {
		fprintf(outputfile, "%s", line);
	    }
	    /* free the "compilation buffer" */
	    regfree(&reg_t);
	    reg_t_freed = True;
	}
	if (False == regex_found) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: There is no PALPlayerRate entry in pconfig.src!\n");
	    returncode = False;
	    goto quit;
	}
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Cannot open pconfig.src or pconfig.qbo!\n");
	returncode = False;
	goto quit;
    }

  quit:
    if (False == reg_t_freed) {
	regfree(&reg_t);
    }
    if (retrievefile) {
	fclose(retrievefile);
    }
    if (savefile) {
	fclose(savefile);
    }
    if (outputfile) {
	fclose(outputfile);
    }
    if (inputfile) {
	fclose(inputfile);
    }

    return returncode;
}


double allypal_modified(const Uns16 player1, const Uns16 player2)
{
    double modifier = 0.0;
    double alliancefactorsum = 0.0;
    double regulator = 0.0;
    Boolean useid = True;


    alliancefactorsum = pqboostconfig.shipfactor + pqboostconfig.planetfactor + pqboostconfig.minefieldfactor + pqboostconfig.combatfactor + pqboostconfig.visionfactor;
    regulator = (alliancefactorsum>1.5 ? alliancefactorsum / 1.5 : 1.0);

    if (PlayersAreAllies(player1, player2)) {
	if (PlayerAllowsAlly(player1, player2, ALLY_SHIPS)) {
	    modifier += pqboostconfig.shipfactor / regulator;
	}
	if (PlayerAllowsAlly(player1, player2, ALLY_PLANETS)) {
	    modifier += pqboostconfig.planetfactor / regulator;
	}
	if (PlayerAllowsAlly(player1, player2, ALLY_MINES)) {
	    modifier += pqboostconfig.minefieldfactor / regulator;
	}
	if (PlayerAllowsAlly(player1, player2, ALLY_COMBAT)) {
	    modifier += pqboostconfig.combatfactor / regulator;
	}
	if (PlayerAllowsAlly(player1, player2, ALLY_VISION)) {
	    modifier += pqboostconfig.visionfactor / regulator;
	}
	find_player(player1, useid)->alliespalportion[player2] = (modifier>1.0 ? 1.0 : modifier);
	return (pow((double) find_player(player2, useid)->battlemass, pqboostconfig.relativepower) * (modifier>1.0 ? 1.0 : modifier));
    } else {
	find_player(player1, useid)->alliespalportion[player2] = 0.0;
	return 0.0;
    }

}


double alliespalsum_modified(const Uns16 racetocheck)
{
    Uns16 i = 0;
    double result = 0.0;


    for (i = 0; i <= RACE_NR; i++) {
	if (playerslist[i].id > 0) {
	    if (PlayerIsActive(playerslist[i].id)) {
		if (racetocheck == playerslist[i].id) {
		    continue;
		}
		result += allypal_modified(racetocheck, playerslist[i].id);
	    }
	}
    }
    return result;
}



/* Main function.*/

int main(int argc, char *argv[])
{
    FILE *statusfile = NULL;
    FILE *errorlogfile = NULL;
    FILE *configfile = NULL;
    FILE *tmpfile = NULL;
    FILE *scorefile = NULL;
    char *errorlog = NULL;
    int c = 0, returncode = 0;
    Boolean phostlibstatus = False;
    Boolean useid = True;
    Uns16 race = 0, ship = 0;
    Uns16 i = 0, j = 0;
    Uns16 phostmajorversion = 0, phostminorversion = 0;
    double mass2 = 0.0;
    Int32 totalmass = 0;
    Int16 phase = 0;
    Int16 x = 0;
    char *messages[21] = { NULL };
    char *message = NULL;
    const char *messagePtr[1];
    time_t now = 0;


    /*
     * Parse the command-line parameters.
     */
    /*
     * Command line usage is:
     *      pqboost [options] [<GameDirectory> [<RootDirectory>]]
     *      options:
     *
     *               -h OR -H     -- get help
     *               -v           -- print version
     *               -e file      -- use 'file' as an error log file
     *               -p phase     -- PQBOOST phase
     */
    while ((c = pgetopt(argc, argv, "hHvp:e:")) != EOF) {
	switch (c) {
	case 'h':
	case 'H':
	case '?':
	    fprintf(errorlogfile ? errorlogfile : stdout,
		    "\nPQBOOST %s (%s)\n", PQBOOST_VERSION, ReleaseDate);
	    fprintf(errorlogfile ? errorlogfile : stdout,
		    "%s\n\n", Notice);
	    fprintf(errorlogfile ? errorlogfile : stdout,
"Usage: pqboost [options] [<GameDirectory> [<RootDirectory>]]\n\
\n\
options:\n\
         -h OR -H     -- prints this help summary and exits\n\
         -v           -- prints program version and exits\n\
         -e file      -- use 'file' as an error log file\n\
         -p phase     -- PQBOOST phase\n");
	    exit(EXIT_SUCCESS);
	case 'v':
	    printf("pqboost %s (%s)\n", PQBOOST_VERSION, ReleaseDate);
	    exit(EXIT_SUCCESS);
	case 'e':
	    if ((errorlog = MemAlloc(strlen(poptarg) + 1)) == NULL) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error: (errorlog = MemAlloc(strlen(poptarg)+1)) failed\n");
		returncode = EXIT_FAILURE;
		goto error;
	    } else {
		strcpy(errorlog, poptarg);
		errorlogfile = fopen(errorlog, "a");
		if (!errorlogfile) {
		    fprintf(errorlogfile ? errorlogfile : stderr,
			    "Error: Cannot open error logfile '%s' (%s)!\n",
			    errorlog, strerror(errno));
		    returncode = EXIT_FAILURE;
		    goto error;
		} else {
		    now = time((time_t *)NULL);
		    fprintf(errorlogfile, "----------------------------------------\n%.24s\n\n", ctime(&now));
		}
	    }
	    break;
	case 'p':
	    phase = (Int16) atoi(poptarg);
	    if ((phase < 1) || (phase > 3)) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error: '-p' parameter not in range [1,3]!\n	   Type '%s -h' for a usage summary.\n",
			argv[0]);
		returncode = EXIT_FAILURE;
		goto error;
	    }
	    break;
	default:
	    ;
	}
    }

    if (poptind < argc) {
	gGameDirectory = argv[poptind++];
    }
    if (poptind < argc) {
	gRootDirectory = argv[poptind++];
    }
    if (poptind < argc) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Too many arguments. First extraneous parameter is '%s'.\n	Type '%s -h' for a usage summary.\n",
		argv[poptind], argv[0]);
	returncode = EXIT_FAILURE;
	goto error;
    }



    /*
     * Initialize the PDK
     */
    InitPHOSTLib();
    phostlibstatus = True;
    InitializeShips();
    if (!ReadGlobalData()) {
	returncode = EXIT_FAILURE;
	goto error;
    }
    if (!ReadHostData()) {
	returncode = EXIT_FAILURE;
	goto error;
    }

    if (TurnNumber() != 0) {
	GameFilesVersion(&phostmajorversion, &phostminorversion);
	if (phostmajorversion < 4) {
	    if ((phostmajorversion != 3) && (phostminorversion != 2)) {
		fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: Unsupported PHOST version %u.%u\n",
		    phostmajorversion, phostminorversion);
		returncode = EXIT_FAILURE;
		goto error;
	    }
	}
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Info: It makes no sense to run PQBOOST in turn 0.\n");
	returncode = EXIT_SUCCESS;
	goto error;
    }

    /*
     * Open the log file.
     */
    if ((2 == phase) || (3 == phase)) {
	gLogFile = OpenOutputFile("pqboost.log", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR | APPEND_MODE);
    } else {
	gLogFile = OpenOutputFile("pqboost.log", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    }
    if (!gLogFile) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Cannot open pqboost.log!\n");
	returncode = EXIT_FAILURE;
	goto error;
    }

    /*
     * Read the configuration file.
     */
    configfile = OpenInputFile("pqboost.cfg", TEXT_MODE | GAME_OR_ROOT_DIR | NO_MISSING_ERROR);
    if (configfile) {
	ConfigFileReader(configfile, "pqboost.cfg", "PQBOOST", True, qboost_assign);
    }

    /*
     * Initialization
     */
    memset(messages, 0, sizeof(messages));
    memset(messagePtr, 0, sizeof(messagePtr));
    message = NULL;

    if (phase < 1 || phase > 3) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Warning: PQBOOST phase not given in commandline, running as phase 1\n");
	phase = 1;
    }

    /*
     * Backup pconfig.src.
     */
    if (1 == phase) {
	if (!(CopyGameFile("pconfig.src", "pconfig.old"))) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: Cannot backup pconfig.src!\n");
	    returncode = EXIT_FAILURE;
	    goto error;
	}
    }

    /*
     * Create temporary file.
     */
    if (!(CopyGameFile("pconfig.src", "pconfig.qbo"))) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Cannot create pconfig.qbo!\n");
	returncode = EXIT_FAILURE;
	goto error;
    }

    /*
     * Allocate memory.
     */
    if ((message = MemAlloc(840)) == NULL) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"(message = MemAlloc(840)) failed\n");
	returncode = EXIT_FAILURE;
	goto error;
    }
    message[0] = '\0';
    for (i = 0; i < 20; i++) {
	if ((messages[i] = MemAlloc(MAXMSGLINE)) == NULL) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "(messages[%i] = MemAlloc(MAXMSGLINE)) failed\n", i);
	    MemFree(message);
	    for (x = i - 1; x >= 0; x--) {
		MemFree(messages[x]);
	    }
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	strcpy(messages[i], "\0");
    }

    /*
     * Let's go!
     */
    Info("\n---------------------------------------------------------------------");
    Info("PQBOOST %s Phase %d", PQBOOST_VERSION, phase);
    Info("PHOST version %u.%u Turn %u", phostmajorversion, phostminorversion, TurnNumber());
#if !defined (pqbsd) && defined (unix)
    Info("Game directory: %s", basename(gGameDirectory));
    Info("Root directory: %s\n", basename(gRootDirectory));
#else
    Info("Game directory: %s", gGameDirectory);
    Info("Root directory: %s\n", gRootDirectory);
#endif
    Info("BeamsRequired       = %u", pqboostconfig.beamsrequired);
    Info("BaysRequired        = %u", pqboostconfig.baysrequired);
    Info("RelativePower       = %3.2f", pqboostconfig.relativepower);
    Info("EnableSecurity      = %s", pqboostconfig.enablesecurity ? "yes" : "no");
    Info("EnableESBandFCB     = %s", pqboostconfig.enableesbandfcb ? "yes" : "no");
    Info("AllianceGameMode    = %s", pqboostconfig.alliancegamemode ? "yes" : "no");
    if (True == pqboostconfig.alliancegamemode) {
	Info("PlanetFactor        = %3.2f", pqboostconfig.planetfactor);
	Info("ShipFactor          = %3.2f", pqboostconfig.shipfactor);
	Info("Minefieldfactor     = %3.2f", pqboostconfig.minefieldfactor);
	Info("CombatFactor        = %3.2f", pqboostconfig.combatfactor);
	Info("VisionFactor        = %3.2f", pqboostconfig.visionfactor);
    }
    if (noranking == pqboostconfig.showranking) {
	Info("ShowRanking         = %s", "no");
    } else if (individual == pqboostconfig.showranking) {
	Info("ShowRanking         = %s", "individual");
    } else if (public == pqboostconfig.showranking) {
	Info("ShowRanking         = %s", "public");
    } else if (verbosepublic == pqboostconfig.showranking) {
	Info("ShowRanking         = %s", "verbosepublic");
    }
    if (noranking != pqboostconfig.showranking) {
	Info("ShowTotalBattlemass = %s", pqboostconfig.showtotalbattlemass ? "yes" : "no");
	Info("RankinglistInterval = %u", pqboostconfig.rankinglistinterval);
    }
    Info("");

    /*
     * Compute battle masses.
     */
    if ((1 == phase) || (2 == phase)) {
	for (ship = 1; ship <= SHIP_NR; ship++) {
	    if (IsShipExist(ship)) {
		if (((ShipTubeNumber(ship) + (ShipBays(ship) ? ShipBays(ship) + gPconfigInfo->ExtraFighterBays[ShipOwner(ship)] : ShipBays(ship))) > pqboostconfig.baysrequired)
		    || (ShipBeamNumber(ship) > pqboostconfig.beamsrequired)) {
		    if (PlayerIsActive(ShipOwner(ship))) {
			if (True == pqboostconfig.enableesbandfcb) {
			    if (True == gPconfigInfo->FedCrewBonus) {
				if (Federation == EffRaceNR(ShipOwner(ship))) {
				    find_player(ShipOwner(ship), useid)->battlemass += 50;
				    totalmass += 50;
				}
			    }
			    if (True == gPconfigInfo->EngShieldBonus) {
				find_player(ShipOwner(ship), useid)->battlemass += (Int32)((double)EngMoneyCost(ShipEngine(ship)) * ((double)(gPconfigInfo->ShieldBonusRate)/100.0));
				totalmass += (Int32)((double)EngMoneyCost(ShipEngine(ship)) * ((double)(gPconfigInfo->ShieldBonusRate)/100.0));
			    }
			}
			find_player(ShipOwner(ship), useid)->battlemass += HullMass(ShipHull(ship));
			totalmass += HullMass(ShipHull(ship));
		    }
		}
	    }
	}

	/*
	 * Compute PALPlayerRates.
	 */
	if (0 == totalmass) {
	    totalmass = 1;
	}
	Info("Total battlemass: %ld", totalmass);
	for (race = 0; race <= RACE_NR; race++) {
	    if (playerslist[race].id > 0) {
		Info("  Player %2u battlemass %ld", race, find_player(playerslist[race].id, useid)->battlemass);
		mass2 += pow((double) find_player(playerslist[race].id, useid)->battlemass, pqboostconfig.relativepower);
	    }
	}
	if (False == pqboostconfig.alliancegamemode) {
	    for (race = 0; race <= RACE_NR; race++) {
		if (playerslist[race].id > 0) {
		    if (PlayerIsActive(playerslist[race].id)) {
			if (mass2 > 0) {
			    find_player(playerslist[race].id, useid)->pal = (Int32) floor(0.5 + 100.0 * (1 - pow((double) find_player(playerslist[race].id, useid)->battlemass, pqboostconfig.relativepower) / mass2));
			    if (find_player(playerslist[race].id, useid)->pal < 0) {
				find_player(playerslist[race].id, useid)->pal = 0;
			    }
			}
		    }
		}
	    }
	} else {
	    for (race = 0; race <= RACE_NR; race++) {
		if (playerslist[race].id > 0) {
		    if (PlayerIsActive(playerslist[race].id)) {
			if (mass2 > 0) {
			    find_player(playerslist[race].id, useid)->pal = (Int32) floor(0.5 + 100.0 * (1 - (pow((double) find_player(playerslist[race].id, useid)->battlemass, pqboostconfig.relativepower) + alliespalsum_modified(playerslist[race].id)) / mass2));
			    if (find_player(playerslist[race].id, useid)->pal < 0) {
				find_player(playerslist[race].id, useid)->pal = 0;
			    }
			}
		    }
		}
	    }
	}

	/*
	 * Print PALPlayerRates.
	 */
/*
	if ((0 == totalmass) && (3 != phase)) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: Total battlemass = 0 in pqboost phase != 3. Program terminated!\n");
	    returncode = EXIT_FAILURE;
	    goto error;
	}
*/
	fprintf(gLogFile, "\nPALPlayerRate = ");
	fprintf(stdout, "\nPALPlayerRate = ");
	for (race = 0; race <= RACE_NR; race++) {
	    if (playerslist[race].id > 0) {
		fprintf(gLogFile, "%ld", find_player(playerslist[race].id, useid)->pal);
		fprintf(stdout, "%ld", find_player(playerslist[race].id, useid)->pal);
		if (race < RACE_NR) {
		    fprintf(gLogFile, ",");
		    fprintf(stdout, ",");
		}
	    }
	}
	fprintf(gLogFile, "\n");
	fprintf(stdout, "\n");
    }

    /*
     *  pconfig.src: Modify PALPlayerRate entry
     */
    if (!(modify_palplayerrate(phase, errorlogfile, pqboostconfig.enablesecurity))) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Cannot modify PALPlayerRate entry in pconfig.src!\n");
	returncode = EXIT_FAILURE;
	goto error;
    }

    /*
     * Send report to players.
     */
    if (2 == phase) {
	if (pqboostconfig.showranking != noranking) {
	    scorefile = OpenOutputFile("pqboost.pts", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
	    if (!scorefile) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error: Cannot open pqboost.pts!\n");
		returncode = EXIT_FAILURE;
		goto error;
	    }
	    /*
	     * Sort the players list by comparing the
	     * battlemass (resp. id) of each player.
	     */
	    qsort(playerslist, count, sizeof(struct player), player_cmp);
	    useid = False;
	    /* Determine ranking list positions. */
	    for (race = 0; race <= RACE_NR; race++) {
		if (race > 0) {
		    if (playerslist[race-1].id > 0) {
			if (playerslist[race-1].battlemass == playerslist[race].battlemass) {
			    playerslist[race].ranking = playerslist[race-1].ranking;
			} else {
			    playerslist[race].ranking = race+1;
			}
		    } else {
			playerslist[race].ranking = race+1;
		    }
		} else {
		    playerslist[race].ranking = race+1;
		}
	    }
	    /*
	     * Sort the players list by comparing the
	     * id of each player.
	     */
	    qsort(playerslist, count, sizeof(struct player), playerfind_cmp);
	    useid = True;
	}
	for (race = 0; race <= RACE_NR; race++) {
	    if (playerslist[race].id > 0) {
		/* Turn 1: Send an info message. */
		if (TurnNumber() == 1) {
		    message[0] = '\0';
		    sprintf(messages[0], "(-a0000)<<< PQBOOST >>>\x0D");
		    strcat(message, messages[0]);
		    sprintf(messages[1], "\x0D");
		    strcat(message, messages[1]);
		    snprintf(messages[2], MAXMSGLINE, "PQBOOST %s\x0D", PQBOOST_VERSION);
		    strcat(message, messages[2]);
		    sprintf(messages[3], "Copyright 1998-2002 Stefan Glasauer.\x0D");
		    strcat(message, messages[3]);
		    sprintf(messages[4], "PDK version by Heiko Schlenker.\x0D");
		    strcat(message, messages[4]);
		    sprintf(messages[5], "PQBOOST comes with ABSOLUTELY NO\x0D");
		    strcat(message, messages[5]);
		    sprintf(messages[6], "WARRANTY. This is free software. See\x0D");
		    strcat(message, messages[6]);
		    sprintf(messages[7], "GNU General Public License for details.\x0D");
		    strcat(message, messages[7]);
		    sprintf(messages[8], "\x0D");
		    strcat(message, messages[8]);
		    sprintf(messages[9], "PQBOOST is a PDK version of QBOOST\x0D");
		    strcat(message, messages[9]);
		    sprintf(messages[10], "which is an add-on to boost the PHOST3\x0D");
		    strcat(message, messages[10]);
		    sprintf(messages[11], "build queue. It modifies PCONFIG.SRC\x0D");
		    strcat(message, messages[11]);
		    sprintf(messages[12], "in order to give weak players better\x0D");
		    strcat(message, messages[12]);
		    sprintf(messages[13], "chances to build ships after the 500\x0D");
		    strcat(message, messages[13]);
		    sprintf(messages[14], "ship limit is reached.\x0D");
		    strcat(message, messages[14]);
		    messagePtr[0] = message;
		    WriteAUXHOSTMessage(race, messagePtr[0]);
		}
		message[0] = '\0';
		sprintf(messages[0], "(-a0000)<<< PQBOOST report >>>\x0D");
		strcat(message, messages[0]);
		sprintf(messages[1], "\x0D");
		strcat(message, messages[1]);
		snprintf(messages[2], MAXMSGLINE, "Turn: %i\x0D", TurnNumber());
		strcat(message, messages[2]);
		sprintf(messages[3], "\x0D");
		strcat(message, messages[3]);

		if (True == pqboostconfig.enablesecurity) {
		    if (False == pqboostconfig.alliancegamemode) {
			snprintf(messages[4], MAXMSGLINE, "Battlemass [kt]: %ld\x0D", find_player(playerslist[race].id, useid)->battlemass);
			strcat(message, messages[4]);
		    } else {
			sprintf(messages[4], "Your effective battlemass includes:\x0D");
			strcat(message, messages[4]);
			snprintf(messages[5], MAXMSGLINE, " 100%s of your battlemass (%ld kt)\x0D", "%", find_player(playerslist[race].id, useid)->battlemass);
			strcat(message, messages[5]);
			for (i = 1; i <= RACE_NR; i++) {
			    if (find_player(playerslist[race].id, useid)->alliespalportion[i] > 0.0)
			    {
				snprintf(messages[i+5], MAXMSGLINE, " %d%s of the battlemass of player %d\x0D", (Uns16) floor(0.5 + 100.0 * find_player(playerslist[race].id, useid)->alliespalportion[i]), "%", i);
				strcat(message, messages[i+5]);
			    }
			}
		    }
		} else {
		    snprintf(messages[4], MAXMSGLINE, "Battlemass (BM), total [kt]: %6ld\x0D", totalmass);
		    strcat(message, messages[4]);
		    snprintf(messages[5], MAXMSGLINE, "Battlemass [kt]:             %6ld\x0D", find_player(playerslist[race].id, useid)->battlemass);
		    strcat(message, messages[5]);
		    snprintf(messages[6], MAXMSGLINE, "PALPlayerRate:               %6ld\x0D", find_player(playerslist[race].id, useid)->pal);
		    strcat(message, messages[6]);
		    if (True == pqboostconfig.alliancegamemode) {
			sprintf(messages[7], "\x0D");
			strcat(message, messages[7]);
			sprintf(messages[8], "Your effective BM includes:\x0D");
			strcat(message, messages[8]);
			snprintf(messages[9], MAXMSGLINE, " 100%s of your BM (%ld kt)\x0D", "%", find_player(playerslist[race].id, useid)->battlemass);
			strcat(message, messages[9]);
			for (i = 1; i <= RACE_NR; i++) {
			    if (find_player(playerslist[race].id, useid)->alliespalportion[i] > 0.0)
			    {
				snprintf(messages[i+9], MAXMSGLINE, " %d%s of the BM of player %d (%ld kt)\x0D", (Uns16) floor(0.5 + 100.0 * find_player(playerslist[race].id, useid)->alliespalportion[i]), "%", i, (Int32) ((find_player(playerslist[race].id, useid)->alliespalportion[i]) * (find_player(playerslist[i].id, useid)->battlemass)));
				strcat(message, messages[i+9]);
			    }
			}
		    }
		}
		messagePtr[0] = message;
		WriteAUXHOSTMessage(race, messagePtr[0]);

		if (pqboostconfig.showranking != noranking) {
		    if (0 == TurnNumber() % pqboostconfig.rankinglistinterval) {
			/*
			 * Sort the players list by comparing the
			 * battlemass (resp. id) of each player.
			 */
			qsort(playerslist, count, sizeof(struct player), player_cmp);
			useid = False;
			if ((public == pqboostconfig.showranking) || (verbosepublic == pqboostconfig.showranking) || (individual == pqboostconfig.showranking)) {
			    message[0] = '\0';
			    sprintf(messages[0], "(-a0000)<<< PQBOOST ranking list >>>\x0D");
			    strcat(message, messages[0]);
			    sprintf(messages[1], "\x0D");
			    strcat(message, messages[1]);
			    snprintf(messages[2], MAXMSGLINE, "Turn: %i\x0D", TurnNumber());
			    strcat(message, messages[2]);
			    sprintf(messages[3], "\x0D");
			    strcat(message, messages[3]);
			    if (public == pqboostconfig.showranking) {
				if (1 == race) {
				    Info("\n\nRanking list according to the battlemass:");
				    fprintf(scorefile, "PQBOOST %s Phase %d\n\n", PQBOOST_VERSION, phase);
				    fprintf(scorefile, "Turn: %i\n\n", TurnNumber());
				    fprintf(scorefile, "Ranking list according to the battlemass:\n");
				}
				if (True == pqboostconfig.showtotalbattlemass) {
				    snprintf(messages[4], MAXMSGLINE, "Battlemass, total [kt]: %6ld\x0D", totalmass);
				    strcat(message, messages[4]);
				    sprintf(messages[5], "\x0D");
				    strcat(message, messages[5]);
				}
				sprintf(messages[6], "Ranking list (battlemass):\x0D");
				strcat(message, messages[6]);
				for (i = 0, j = 1; i <= RACE_NR; i++) {
				    if (playerslist[i].id > 0) {
					if (1 == race) {
					    if (i > 0) {
						if (playerslist[i-1].id > 0) {
						    if (playerslist[i-1].battlemass == playerslist[i].battlemass) {
							Info("    player %2u", playerslist[i].id);
							fprintf(scorefile, "    player %2u\n", playerslist[i].id);
						    } else {
							Info("%2u. player %2u", j, playerslist[i].id);
							fprintf(scorefile, "%2u. player %2u\n", j, playerslist[i].id);
						    }
						} else {
						    Info("%2u. player %2u", j, playerslist[i].id);
						    fprintf(scorefile, "%2u. player %2u\n", j, playerslist[i].id);
						}
					    } else {
						Info("%2u. player %2u", j, playerslist[i].id);
						fprintf(scorefile, "%2u. player %2u\n", j, playerslist[i].id);
					    }
					}
					if (i > 0) {
					    if (playerslist[i-1].id > 0) {
						if (playerslist[i-1].battlemass == playerslist[i].battlemass) {
						    snprintf(messages[6+j], MAXMSGLINE, "    player %2u\x0D", playerslist[i].id);
						    strcat(message, messages[6+j]);
						} else {
						    snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u\x0D", j, playerslist[i].id);
						    strcat(message, messages[6+j]);
						}
					    } else {
						snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u\x0D", j, playerslist[i].id);
						strcat(message, messages[6+j]);
					    }
					} else {
					    snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u\x0D", j, playerslist[i].id);
					    strcat(message, messages[6+j]);
					}
					j++;
				    }
				}
			    } else if (verbosepublic == pqboostconfig.showranking) {
				if (1 == race) {
				    Info("\n\nRanking list according to the battlemass:");
				    fprintf(scorefile, "PQBOOST %s Phase %d\n\n", PQBOOST_VERSION, phase);
				    fprintf(scorefile, "Turn: %i\n\n", TurnNumber());
				    fprintf(scorefile, "Ranking list according to the battlemass:\n");
				}
				if (True == pqboostconfig.showtotalbattlemass) {
				    snprintf(messages[4], MAXMSGLINE, "Battlemass, total [kt]: %6ld\x0D", totalmass);
				    strcat(message, messages[4]);
				    sprintf(messages[5], "\x0D");
				    strcat(message, messages[5]);
				}
				sprintf(messages[6], "Ranking list (battlemass):\x0D");
				strcat(message, messages[6]);
				for (i = 0, j = 1; i <= RACE_NR; i++) {
				    if (playerslist[i].id > 0) {
					if (1 == race) {
					    if (i > 0) {
						if (playerslist[i-1].id > 0) {
						    if (playerslist[i-1].battlemass == playerslist[i].battlemass) {
							Info("    player %2u: %6ld kt", playerslist[i].id, playerslist[i].battlemass);
							fprintf(scorefile, "    player %2u: %6ld kt\n", playerslist[i].id, playerslist[i].battlemass);
						    } else {
							Info("%2u. player %2u: %6ld kt", j, playerslist[i].id, playerslist[i].battlemass);
							fprintf(scorefile, "%2u. player %2u: %6ld kt\n", j, playerslist[i].id, playerslist[i].battlemass);
						    }
						} else {
						    Info("%2u. player %2u: %6ld kt", j, playerslist[i].id, playerslist[i].battlemass);
						    fprintf(scorefile, "%2u. player %2u: %6ld kt\n", j, playerslist[i].id, playerslist[i].battlemass);
						}
					    } else {
						Info("%2u. player %2u: %6ld kt", j, playerslist[i].id, playerslist[i].battlemass);
						fprintf(scorefile, "%2u. player %2u: %6ld kt\n", j, playerslist[i].id, playerslist[i].battlemass);
					    }
					}
					if (i > 0) {
					    if (playerslist[i-1].id > 0) {
						if (playerslist[i-1].battlemass == playerslist[i].battlemass) {
						    snprintf(messages[6+j], MAXMSGLINE, "    player %2u: %6ld kt\x0D", playerslist[i].id, playerslist[i].battlemass);
						    strcat(message, messages[6+j]);
						} else {
						    snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u: %6ld kt\x0D", j, playerslist[i].id, playerslist[i].battlemass);
						    strcat(message, messages[6+j]);
						}
					    } else {
						snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u: %6ld kt\x0D", j, playerslist[i].id, playerslist[i].battlemass);
						strcat(message, messages[6+j]);
					    }
					} else {
					    snprintf(messages[6+j], MAXMSGLINE, "%2u. player %2u: %6ld kt\x0D", j, playerslist[i].id, playerslist[i].battlemass);
					    strcat(message, messages[6+j]);
					}
					j++;
				    }
				}
			    } else if (individual == pqboostconfig.showranking) {
				qsort(playerslist, count, sizeof(struct player), playerfind_cmp);
				useid = True;
				if (True == pqboostconfig.showtotalbattlemass) {
				    snprintf(messages[4], MAXMSGLINE, "Battlemass, total [kt]: %6ld\x0D", totalmass);
				    strcat(message, messages[4]);
				}
				snprintf(messages[5], MAXMSGLINE, "Battlemass [kt]:        %6ld\x0D", find_player(playerslist[race].id, useid)->battlemass);
				strcat(message, messages[5]);
				snprintf(messages[6], MAXMSGLINE, "Ranking list position:  %6d\x0D", find_player(playerslist[race].id, useid)->ranking);
				strcat(message, messages[6]);
			    }
			    messagePtr[0] = message;
			    WriteAUXHOSTMessage(race, messagePtr[0]);
			}
			/*
			 * Sort the players list by comparing the
			 * id of each player.
			 */
			if (False == useid) {
			    qsort(playerslist, count, sizeof(struct player), playerfind_cmp);
			    useid = True;
			}
		    }
		}
	    }
	}
    }

    /*
     * Remove temporary files.
     */
    if ((3 == phase) && (True == pqboostconfig.enablesecurity)) {
	if (!(RemoveGameFile("pqboost.tmp"))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "Error: Cannot remove pqboost.tmp. Possible error: PQBOOST phase 1 not executed!\n");
	    returncode = EXIT_FAILURE;
	    goto error;
	}
    }
    tmpfile = OpenInputFile("pconfig.qbo", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (tmpfile) {
	fclose(tmpfile);
	tmpfile = NULL;
	if (!(RemoveGameFile("pconfig.qbo"))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "Error: Cannot remove pconfig.qbo!\n");
	    returncode = EXIT_FAILURE;
	    goto error;
	}
    }

    /*
     * Close down.
     */
error:

    if (scorefile) {
	fclose(scorefile);
    }
    if (configfile) {
	fclose(configfile);
    }
    if (gLogFile) {
	fclose(gLogFile);
    }
    if (errorlogfile) {
	if (returncode != EXIT_SUCCESS) {
	    fprintf(stderr, "ERROR! Please read the error log file '%s'.\n", errorlog);
	}
	if (3 == phase) {
	    fprintf(errorlogfile, "----------------------------------------\n########################################\n");
	}
	fclose(errorlogfile);
	MemFree(errorlog);
    }
    if (message) {
	for (i = 0; i < 20; i++) {
	    MemFree(messages[i]);
	}
	MemFree(message);
    }
    if (statusfile) {
	fclose(statusfile);
    }
    if (True == phostlibstatus) {
	FreePHOSTLib();
    }

    exit(returncode);
}
